# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import numpy as np
import scipy as sp


def gauss_quadrature_abscissa_weight(
    n_points: int,
    g_kind: str = "Legendre",
    alpha: float = 0.0,  # only for "Laguerre"
) -> tuple[np.ndarray, np.ndarray]:
    #
    # g_kind
    # "Chebyshev1" : x = [-1, 1],     w(x) = (1-x^2)^{-1/2}, (or Chebyshev)
    # "Chebyshev2" : x = [-1, 1],     w(x) = (1-x^2)^{1/2}
    # "Legendre"   : x = [-1, 1],     w(x) = 1
    # "Lobatto"    : x = [-1, 1],     w(x) = 1
    # "Laguerre"   : x = [0, inf],    w(x) = x^\alpha exp(-x)
    # "Hermite"    : x = [-inf, inf], w(x) = exp(-x^2)
    #
    # If you use abscissa and weight only in python, better to use these.
    #   abscissa, weight = np.polynomial.chebyshev.chebgauss(n_points) # Chebyshev of the first kind
    #   abscissa, weight = np.polynomial.legendre.leggauss(n_points) # Legendre
    #   abscissa, weight = np.polynomial.laguerre.laggauss(n_points) # Laguerre for alpha=0
    #   abscissa, weight = np.polynomial.hermite.hermgauss(n_points) # Hermite
    #
    if n_points <= 0:
        # Error
        raise ValueError(f"Unexpected component: Gauss-{g_kind} quadrature with n_points={n_points}")

    # Obtain abscissa and weight for Gauss-quadrature using Golub-Welsh algorithm
    #  (G.H. Golub, J.H. Welsch, "Calculation of Gauss quadrature rules", Math. Comp., 23 (1969), pp. 221-230)

    # "Legendre", "Lobatto", "Radau", "InvRadau": x=[-1, 1]
    if g_kind != "Lobatto":
        N = n_points
        a = np.zeros(N, dtype=np.float64)
        b = np.zeros(N, dtype=np.float64)
        c = np.zeros(N, dtype=np.float64)
        d = np.zeros(N, dtype=np.float64)
        h = np.zeros(N, dtype=np.float64)

        if g_kind == "Chebyshev" or g_kind == "Chebyshev1":
            k0 = 1.0  # 22.3.6 n=0 of k_n in A&S, with difference for first term (x2)
            for n in range(N):
                a[n] = 1.0  # 22.7.4, a_{1,n} in A&S
                b[n] = 0.0  # 22.7.4, a_{2,n} in A&S
                d[n] = 1.0  # 22.7.4, a_{4,n} in A&S
            c[0] = 1.0
            for n in range(1, N):
                c[n] = 2.0  # 22.7.4, a_{3,n} in A&S
            h[0] = np.pi
            for n in range(1, N):
                h[n] = np.pi * 0.5  # 22.2.4 h_n in A&S
        elif g_kind == "Chebyshev2":
            k0 = 1.0  # 22.3.7 n=0 of k_n in A&S
            for n in range(N):
                a[n] = 1.0  # 22.7.5, a_{1,n} in A&S
                b[n] = 0.0  # 22.7.5, a_{2,n} in A&S
                c[n] = 2.0  # 22.7.5, a_{3,n} in A&S
                d[n] = 1.0  # 22.7.5, a_{4,n} in A&S
                h[n] = np.pi * 0.5  # 22.2.5 h_n in A&S
        elif g_kind == "Legendre":
            k0 = 1.0  # 22.3.8 n=0 of k_n in A&S
            for n in range(N):
                a[n] = n + 1.0  # 22.7.10, a_{1,n} in A&S
                b[n] = 0.0  # 22.7.10, a_{2,n} in A&S
                c[n] = 2.0 * n + 1.0  # 22.7.10, a_{3,n} in A&S
                d[n] = n * 1.0  # 22.7.10, a_{4,n} in A&S
                h[n] = 2.0 / (2.0 * n + 1.0)  # 22.2.10 h_n in A&S
        elif g_kind == "Laguerre":
            k0 = 1.0  # 22.3.9 n=0 of k_n in A&S
            for n in range(N):
                a[n] = n + 1.0  # 22.7.12, a_{1,n} in A&S
                b[n] = 2.0 * n + alpha + 1.0  # 22.7.12, a_{2,n} in A&S
                c[n] = -1.0  # 22.7.12, a_{3,n} in A&S
                d[n] = n + alpha  # 22.7.12, a_{4,n} in A&S
            h[0] = sp.special.gamma(alpha + 1.0)
            for n in range(1, N):
                h[n] = ((alpha + n) / n) * h[n - 1]  # 22.2.13 h_n in A&S
        elif g_kind == "Hermite":
            k0 = 1.0  # 22.3.10 n=0 of k_n in A&S
            for n in range(N):
                a[n] = 1.0  # 22.7.13, a_{1,n} in A&S
                b[n] = 0.0  # 22.7.13, a_{2,n} in A&S
                c[n] = 2.0  # 22.7.13, a_{3,n} in A&S
                d[n] = 2.0 * n  # 22.7.13, a_{4,n} in A&S
            h[0] = np.sqrt(np.pi)
            for n in range(1, N):
                h[n] = 2.0 * n * h[n - 1]  # 22.2.14 h_n in A&S
        #    elif g_kind == "Jacobi":
        #        k0 = 1.0  # 22.3.1 n=0 of k_n in A&S
        #        for n in range(N):
        #            a[n] = 2.0 * (n + 1.0) * (n + alpha + beta + 1.0) * (2.0 * n + alpha + beta)  # 22.7.1, a_{1,n} in A&S
        #            b[n] = (2.0 * n + alpha + beta + 1.0) * (alpha + beta) * (alpha - beta)  # 22.7.1, a_{2,n} in A&S
        #            c[n] = (2.0 * n + alpha + beta) * (2.0 * n + alpha + beta + 1.0) * (2.0 * n + alpha + beta + 2.0)  # 22.7.1, a_{3,n} in A&S
        #            d[n] = 2.0 * (n + alpha) * (n + beta) * (2.0 * n + alpha + beta + 2.0)  # 22.7.1, a_{4,n} in A&S
        #        ga = sp.special.gamma(alpha + 1.0)
        #        gb = sp.special.gamma(beta + 1.0)
        #        gab = sp.special.gamma(alpha + beta + 1.0)
        #        h[0] = ((2 ** (alpha + beta + 1.0)) / (alpha + beta + 1.0)) * ga * gb / gab
        #        for n in range(1, N):
        #            c1 = (2.0 * n + alpha + beta - 1.0) / (2.0 * n + alpha + beta + 1.0)
        #            c2 = (n + alpha) * (n + beta) / (n * (n + alpha + beta))
        #            h[n] = c1 * c2 * h[n - 1]  # 22.2.1 h_n in A&S

        # Obtain symmetric a tridiagonal matrix elements
        diag = np.zeros(N, dtype=np.float64)  # alpha
        off_diag = np.zeros(N - 1, dtype=np.float64)  # beta
        for n in range(N - 1):
            diag[n] = -b[n] / c[n]
            off_diag[n] = np.sqrt((a[n] * d[n + 1]) / (c[n] * c[n + 1]))
        diag[N - 1] = -b[N - 1] / c[N - 1]

        kappa = np.zeros(N, dtype=np.float64)
        kappa[0] = 1.0
        for n in range(1, N):
            kappa[n] = np.sqrt((a[n - 1] * c[n]) / (c[n - 1] * d[n])) * kappa[n - 1]

        # Diagonalize a tridiagonal matrix
        abscissa = np.zeros(N, dtype=np.float64)
        weight = np.zeros(N, dtype=np.float64)
        abscissa, eigenvectors = sp.linalg.eigh_tridiagonal(diag, off_diag, eigvals_only=False)

        for j in range(N):
            # q'_n to q_n for all n = 0, 1, ..., N-1
            eigenvectors[0:N, j] = eigenvectors[0:N, j] / kappa[0:N]
            q0j = eigenvectors[0, j]
            # q_n to p_n for all n = 0, 1, ..., N-1 where p_n is the orthfgonal polynomials at x=x_j, (j = 0, 1, ..., N-1)
            eigenvectors[0:N, j] = eigenvectors[0:N, j] * k0 / q0j

        for j in range(N):
            s = 0.0
            for n in range(N):
                s = s + eigenvectors[n, j] ** 2 / h[n]
            weight[j] = 1.0 / s

    elif g_kind == "Lobatto":
        if n_points <= 1:
            # Error
            raise ValueError(f"Unexpected component: Gauss-{g_kind} quadrature with n_points={n_points}")
        elif n_points == 2:
            abscissa = np.zeros(n_points, dtype=np.float64)
            weight = np.zeros(n_points, dtype=np.float64)
            abscissa[0] = -1.0
            abscissa[1] = 1.0
            weight[0] = 1.0
            weight[1] = 1.0
            return abscissa, weight
        else:
            N = n_points - 2
            a = np.zeros(N, dtype=np.float64)
            b = np.zeros(N, dtype=np.float64)
            c = np.zeros(N, dtype=np.float64)
            d = np.zeros(N, dtype=np.float64)
            h = np.zeros(N, dtype=np.float64)

            k0 = -1.5
            for n in range(N):
                a[n] = (n + 3.0) / (2.0 * n + 5.0)
                b[n] = 0.0
                c[n] = 1.0
                d[n] = n / (2.0 * n + 1.0)
                h[n] = (2.0 * (2.0 * n + 3.0)) / ((n + 2.0) * (n + 1.0))

            # Obtain symmetric a tridiagonal matrix elements
            diag = np.zeros(N, dtype=np.float64)  # alpha
            off_diag = np.zeros(N - 1, dtype=np.float64)  # beta
            for n in range(N - 1):
                diag[n] = -b[n] / c[n]
                off_diag[n] = np.sqrt((a[n] * d[n + 1]) / (c[n] * c[n + 1]))
            diag[N - 1] = -b[N - 1] / c[N - 1]

            kappa = np.zeros(N, dtype=np.float64)
            kappa[0] = 1.0
            for n in range(1, N):
                kappa[n] = np.sqrt((a[n - 1] * c[n]) / (c[n - 1] * d[n])) * kappa[n - 1]

            # Diagonalize a tridiagonal matrix
            x_nodes = np.zeros(N, dtype=np.float64)
            p_x_nodes = np.zeros(N, dtype=np.float64)
            x_nodes, p_x_nodes = sp.linalg.eigh_tridiagonal(diag, off_diag, eigvals_only=False)

            for j in range(N):
                # q'_n to q_n for all n = 0, 1, ..., N-1
                p_x_nodes[0:N, j] = p_x_nodes[0:N, j] / kappa[0:N]
                q0j = p_x_nodes[0, j]
                # q_n to p_n for all n = 0, 1, ..., N-1 where p_n is the orthfgonal polynomials at x=x_j, (j = 0, 1, ..., N-1)
                p_x_nodes[0:N, j] = p_x_nodes[0:N, j] * k0 / q0j

            w_nodes = np.zeros(N, dtype=np.float64)
            for j in range(N):
                s = 0.0
                for n in range(N):
                    s = s + p_x_nodes[n, j] ** 2 / h[n]
                w_nodes[j] = 1.0 / s
            # Additional factor for Gauss-Lobatto
            w_nodes /= (1 - x_nodes[0:N]) * (1 + x_nodes[0:N])

            # Balance of the array size, namely include the edge points
            abscissa = np.zeros(n_points, dtype=np.float64)
            weight = np.zeros(n_points, dtype=np.float64)

            w_0N = 2.0 / (n_points * (n_points - 1.0))
            abscissa[0] = -1.0
            abscissa[1 : n_points - 1] = x_nodes[0 : n_points - 2]
            abscissa[n_points - 1] = 1.0
            weight[0] = w_0N
            weight[1 : n_points - 1] = w_nodes[0 : n_points - 2]
            weight[n_points - 1] = w_0N

    else:
        raise ValueError(f"Undefined quadrature rule: your g_kind = {g_kind}")

    return abscissa, weight


def func(x):
    f = x**4
    return f


if __name__ == "__main__":
    # Input parameters

    n_points = 4

    # Assuming integration: \int_{a}^{b} ((x-a)(b-x))^{-1/2} f(x) dx
    x, w = gauss_quadrature_abscissa_weight(n_points, "Chebyshev1")
    a, b = -1.0, 1.0  # Scaling: x=[a, b]
    w = w
    x = (x * (b - a) + (b + a)) * 0.5
    s = 0.0
    for i in range(n_points):
        s += w[i] * func(x[i])
    print("Chebyshev1", s)

    # Assuming integration: \int_{a}^{b} ((x-a)(b-x))^{1/2} f(x) dx
    x, w = gauss_quadrature_abscissa_weight(n_points, "Chebyshev2")
    a, b = -1.0, 1.0  # Scaling: x=[a, b]
    w = w * ((b - a) * 0.5) ** 2
    x = (x * (b - a) + (b + a)) * 0.5
    s = 0.0
    for i in range(n_points):
        s += w[i] * func(x[i])
    print("Chebyshev2", s)

    # Assuming integration: \int_{a}^{b} f(x) dx
    x, w = gauss_quadrature_abscissa_weight(n_points, "Legendre")
    a, b = -1.0, 1.0  # Scaling: x=[a, b]
    w = w * (b - a) * 0.5
    x = (x * (b - a) + (b + a)) * 0.5
    s = 0.0
    for i in range(n_points):
        s += w[i] * func(x[i])
    print("Legendre", s)

    # Assuming integration: \int_{a}^{b} f(x) dx
    x, w = gauss_quadrature_abscissa_weight(n_points, "Lobatto")
    a, b = -1.0, 1.0  # Scaling: x=[a, b]
    w = w * (b - a) * 0.5
    x = (x * (b - a) + (b + a)) * 0.5
    s = 0.0
    for i in range(n_points):
        s += w[i] * func(x[i])
    print("Lobatto", s)

    # Assuming integration: \int_0^\infty x^\alpha e^{-\kappa*x} f(x) dx
    alpha = -0.25  # > -1
    x, w = gauss_quadrature_abscissa_weight(n_points, "Laguerre", alpha=alpha)
    kappa = 2.0  # Scaling
    w = w / (kappa ** (alpha + 1.0))
    x = x / kappa
    s = 0.0
    for i in range(n_points):
        s += w[i] * func(x[i])
    print("Laguerre", s)

    # Assuming integration: \int_{-\infty}^\infty e^{-\kappa*x^2} f(x) dx
    x, w = gauss_quadrature_abscissa_weight(n_points, "Hermite")
    kappa = 2.0
    w = w / np.sqrt(kappa)
    x = x / np.sqrt(kappa)
    s = 0.0
    for i in range(n_points):
        s += w[i] * func(x[i])
    print("Hermite", s)
